// 				allocate.cc
//
//  written by Zheng Gengbin

/*
This file contains passes :
	1. split_allocate
	   split allocate statement as allocate one array a time.
	2. deal_allocate
	   insert align or distribute statement after allocate statement as
	   re-align and re-distribute.
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "strategy.h"

#include "pass.h"
#include "ext_func.h"


/****************************************************************************/

// split allocate(a,b)
// as allocate(a) : allocate(b)
void split_allocate(SgStatement *s)
{
   for(SgExpression * expr=s->expr(0); expr; expr=expr->rhs())  
   {
        SgExpression *exp = expr->lhs();
	SgExpression *newexpr = new SgExpression(EXPR_LIST, exp->copyPtr(), NULL, NULL);
	SgStatement *alloc = new SgStatement(ALLOCATE_STMT, NULL, NULL, newexpr, NULL, NULL);
	s->insertStmtBefore(*alloc, *s->controlParent());
   }
   s->deleteStmt();
}

/****************************************************************************/

// change lbound call in align statement
static void change_lbound_by_allocate(SgStatement *realign_stmt, SgStatement *alloc_stmt)
{
   SgExpression *expr = realign_stmt->expr(2);
   SgSymbol *lbound = alloc_stmt->getSymbolInCurrentScope("lbound");
   if (!expr->IsSymbolInExpression(*lbound)) return;
   SgArrayRefExp *array = isSgArrayRefExp(alloc_stmt->expr(0)->lhs());
   ASSERT(array);
   SgExpression *func;
   while (func = expr->IsSymbolInExpression(*lbound))
   {
	SgIntrinsicFunctionCallExp *lb = isSgIntrinsicFunctionCallExp(func);
	SgExpression *dim = lb->arg(1);
	ASSERT(dim->isInteger());
	int d = dim->valueInteger() - 1;
	SgExpression *l = get_lbound_from_arrayexp(array, d);
	ASSERT(l);
	func->replace(*l);
   }
}

//  move distributee and alignee after allocate
//  change them as re-distribute and re-align executable statement
void deal_allocate(SgStatement *hedr_stmt)
{
	SgSymbol * symbol = hedr_stmt->symbol()->next();
	SgSymbol * sym_son , *sym_par;

	while ((symbol!=NULL)) {
    		if (symbol->scope()==hedr_stmt)
      			break;
    		symbol=symbol->next();
  	}
	if (symbol==NULL) return ;

	for( SgSymbol *sym=symbol ; sym; sym=sym->nextInScope()) 
	{
	    if (sym->scope() == NULL ) continue;
	    if ( (sym->attributes() &  ALLOCATABLE_BIT) == 0 ) continue;

	      // for allocatable symbol , add redistribute stmt after allocate statements
	    if (sym->numberOfAttributes(Dist_Stmt) > 0)
            {
		SgStatement *dist_stmt = (SgStatement *)sym->attributeValue(0, Dist_Stmt);
	    	for (int j=0 ; j< sym->numberOfAttributes(Alloc_Stmt) ; j++)
	    	{
		SgStatement *alloc_stmt=(SgStatement*)sym->attributeValue(j,Alloc_Stmt);
		SgStatement *redist_stmt = dist_stmt->copyPtr();
	        redist_stmt->thebif->variant = REDISTRIBUTE_STMT;
        	alloc_stmt->insertStmtAfter(*redist_stmt,*alloc_stmt->controlParent());
	    	}
		dist_stmt->deleteStmt();
            }
	      // for allocatable symbol , add realign stmt after allocate statements
	    if (sym->numberOfAttributes(Align_Stmt) > 0)
            {
		SgStatement *align_stmt = (SgStatement *)sym->attributeValue(0, Align_Stmt);
	    	for (int j=0 ; j< sym->numberOfAttributes(Alloc_Stmt) ; j++)
	    	{
		SgStatement *alloc_stmt=(SgStatement*)sym->attributeValue(j,Alloc_Stmt);
		  // align_stmt may have lbound func call.
		SgStatement *realign_stmt = align_stmt->copyPtr();
		change_lbound_by_allocate(realign_stmt, alloc_stmt);
	        realign_stmt->thebif->variant = REALIGN_STMT;
        	alloc_stmt->insertStmtAfter(*realign_stmt,*alloc_stmt->controlParent());
	    	}
		align_stmt->deleteStmt();
            }
	      
	}
}
